/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2006-2009 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 */

/*
 *  This file contains a driver for the watchdog timers.
 *  We use them in "timer mode", where a counter decreases
 *  until reaching 0, generating an interrupt (number 30).
 *  If we put them in "watchdog mode" they would reset the
 *  CPU!!
 */
#include "kernel_impl.h"
#include "time_event.h"
#include <sil.h>
#include "target_timer.h"

/*
 *  Clear the timer interrupt request
 */
Inline void
target_timer_int_clear()
{
	/* clear pending bit */
	sil_wrw_mem((void *)MPCORE_WDT_ISR, MPCORE_WDT_ISR_SCBIT);
}

/*
 *  Timer initialization
 */
void
target_timer_initialize(intptr_t exinf)
{
	uint32_t cyc = TO_CLOCK(TIC_NUME, TIC_DENO);

	/* Stop the timer */
	sil_wrw_mem((void *)MPCORE_WDT_CNT, 0x00);

	/* Set it to timer mode */
	sil_wrw_mem((void *)MPCORE_WDT_DR, 0x12345678);
	sil_wrw_mem((void *)MPCORE_WDT_DR, 0x87654321);

	/* Stop the timer */
	sil_wrw_mem((void *)MPCORE_WDT_CNT, 0x00);

	/* Clear pending interrupts */
	target_timer_int_clear();

	/* Set the counter value (1ms per tick) */
	sil_wrw_mem((void *)MPCORE_WDT_LR, cyc - 1);

	/*
	 * Set the scaler for 1ms per tick and start the timer
	 */
	sil_wrw_mem((void *)MPCORE_WDT_CNT,
				(MPCORE_WDT_PS_1MS << MPCORE_WDT_CNT_PS_OFFSET) |
				MPCORE_WDT_CNT_IEN | MPCORE_WDT_CNT_AR | MPCORE_WDT_CNT_EN);
}

/*
 *  Timer termination
 */
void
target_timer_terminate(intptr_t exinf)
{
	/* Clear interrupts */
	target_timer_int_clear();

	/* Stop the timer */
	sil_wrw_mem((void *)MPCORE_WDT_CNT, 0x00);
}

/*
 *  Timer interrupt handler
 */
void
target_timer_handler(void)
{
	ID prcid;

	/* Clear interrupts */
	target_timer_int_clear();

	/* Signal the system tick */
	iget_pid(&prcid);
	i_begin_int((0x10000 << (prcid - 1))|DIC_IRQNO_WDT);
	signal_time();
	i_end_int((0x10000 << (prcid - 1))|DIC_IRQNO_WDT);
}
