/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2005-2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 */
#ifndef TOPPERS_CHIP_TIMER_H
#define TOPPERS_CHIP_TIMER_H

#include <sil.h>
#include "mpcore.h"

/*
 * Watchdog Timer Registers
 */
#define MPCORE_WDT_LR     (MPCORE_PMR_BASE + 0x0620)  /* Load register */
#define MPCORE_WDT_COUNT  (MPCORE_PMR_BASE + 0x0624)  /* Counter register */
#define MPCORE_WDT_CNT    (MPCORE_PMR_BASE + 0x0628)  /* Control register */
#define         MPCORE_WDT_CNT_EN        0x1
#define         MPCORE_WDT_CNT_AR        0x2
#define         MPCORE_WDT_CNT_IEN       0x4
#define         MPCORE_WDT_CNT_PS_OFFSET 0x8
#define MPCORE_WDT_ISR    (MPCORE_PMR_BASE + 0x062C)  /* Interrupt status register */
#define         MPCORE_WDT_ISR_SCBIT    0x01
#define MPCORE_WDT_RSR    (MPCORE_PMR_BASE + 0x0630)  /* Reset state register */
#define MPCORE_WDT_DR     (MPCORE_PMR_BASE + 0x0634)  /* Watchdog disable Register */

/*
 *  Timer interrupts number, priority and attributes
 */
#define INHNO_TIMER_PRC1     (0x00010000 | DIC_IRQNO_WDT)
#define INHNO_TIMER_PRC2     (0x00020000 | DIC_IRQNO_WDT)

#define INTNO_TIMER_PRC1     (0x00010000 | DIC_IRQNO_WDT)
#define INTNO_TIMER_PRC2     (0x00020000 | DIC_IRQNO_WDT)

#define INTATR_TIMER   0U

#ifndef TOPPERS_MACRO_ONLY

/*
 *  Type definition for the internal representation of the timer value
 */
typedef uint32_t CLOCK;

/*
 *  Convertion between the internal representation of the timer value
 *  and miliseconds/microseconds. One tick equals to 1ms.
 */
#define TIMER_CLOCK          MPCORE_WDT_LOAD_1MS // cycles
#define TO_CLOCK(nume, deno) (TIMER_CLOCK * (nume) / (deno))
#define TO_USEC(clock)       (((SYSUTM) clock) * 1000U / TIMER_CLOCK)

/*
 *  Maximum timer value (internal representation in cycles)
 */
#define MAX_CLOCK        ((CLOCK) 0xffffffffU)

/*
 *  Approximate amount of time since the timer interrupt occurred
 */
#define GET_TOLERANCE    100U    /* delay tolerance (internal represent.) */
#define BEFORE_IREQ(clock) \
((clock) >= TO_CLOCK(TIC_NUME, TIC_DENO) - GET_TOLERANCE)

/*
 *  Timer initialization
 */
extern void
target_timer_initialize(intptr_t exinf);

/*
 * Timer termination
 */
extern void
target_timer_terminate(intptr_t exinf);

/*
 *  Get the current timer value
 */
Inline CLOCK
target_timer_get_current(void)
{
    return(TO_CLOCK(TIC_NUME, TIC_DENO) -
           sil_rew_mem((void *)MPCORE_WDT_COUNT));
}

/*
 *  Check for pending timer interrupts
 */
Inline bool_t
target_timer_probe_int(void)
{
    if((sil_rew_mem((void *)MPCORE_WDT_ISR) &
        MPCORE_WDT_ISR_SCBIT) == MPCORE_WDT_ISR_SCBIT) {
          return true;
    } else {
        return false;
    }
}

/*
 *  Timer interrupt handler
 */
extern void    target_timer_handler(void);

#endif /* TOPPERS_MACRO_ONLY */
#endif /* TOPPERS_CHIP_TIMER_H */
