/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 * 
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 * 
 *  @(#) $Id: pl310.c 1058 2014-01-10 05:08:11Z ertl-honda $
 */

#include "pl310.h"

/* Bitmask of active ways */
static uint32_t pl310_way_mask;  

static inline void cache_wait_way(uint32_t *reg, unsigned long mask)
{
	/* wait for cache operation by line or way to complete */
	while ( sil_rew_mem((void*)(reg)) & mask );
}

static inline void cache_sync(void)
{
	sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_CACHE_SYNC), 0);
}

static inline void pl310_inv_all(void)
{
	/* invalidate all ways */
	sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_INV_WAY), pl310_way_mask);
	cache_wait_way((uint32_t *)(PL310_L2CACHE_BASE+PL310_INV_WAY), pl310_way_mask);
	cache_sync();
}

void
pl310_init(uint32_t aux_val, uint32_t aux_mask)
{
	uint32_t tmp;
	uint32_t aux;
	uint32_t cache_id;
	uint32_t prefetch;
	uint32_t prefetch_val = 0;    
	uint32_t power;
	int ways;

	/* L2å夬Ǥ˥ˤʤäƤ뤫ǧ */
	tmp = sil_rew_mem((void*)(PL310_L2CACHE_BASE + PL310_CTRL));

	/* L2å夬̵ξΤ߽»ܤ */
	if ( !(tmp & 1) ) {
		cache_id = sil_rew_mem((void*)(PL310_L2CACHE_BASE + PL310_CACHE_ID));
		aux = sil_rew_mem((void*)(PL310_L2CACHE_BASE + PL310_AUX_CTRL));
		prefetch = sil_rew_mem((void*)(PL310_L2CACHE_BASE + PL310_PREFETCH_CTRL));
		power = sil_rew_mem((void*)(PL310_L2CACHE_BASE + PL310_POWER_CTRL));

		if (aux & (1 << 16)) {
			ways = 16;
		}
		else {
			ways = 8;
		}

		aux_val |= 1 << 22;
        
		aux_val |= 1 << 29;
		prefetch_val |= 1 << 29;

		aux_val |= 1 << 28;
		prefetch_val |= 1 << 28;


		if ((cache_id & 0x3f) > 0x6) {
			prefetch_val |= 1 << 30;
		}
                
		pl310_way_mask = (1 << ways) - 1;

		aux &= aux_mask;
		aux |= aux_val;
		prefetch |= prefetch_val;

		sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_AUX_CTRL), aux);
		sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_PREFETCH_CTRL), prefetch);
		sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_POWER_CTRL), power);
        
		pl310_inv_all();
        
		sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_CTRL), 1);
	}    
}

void
pl310_debug_set(uint32_t val)
{
	sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_DEBUG_CTRL), val);
}

void
pl310_flush_all(void)
{
	pl310_debug_set(0x03);
	sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_CLEAN_INV_WAY), pl310_way_mask);
	cache_wait_way((void*)(PL310_L2CACHE_BASE + PL310_CLEAN_INV_WAY), pl310_way_mask);
	cache_sync();
	pl310_debug_set(0x00);
}

void
pl310_disable(void)
{
	pl310_flush_all();
	sil_wrw_mem((void*)(PL310_L2CACHE_BASE + PL310_CTRL), 0x0);
}
