/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2006-2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 */
#ifndef TOPPERS_SOCFPGA_H
#define TOPPERS_SOCFPGA_H

#include <sil.h>

/*
 * CPU clocks
 */
#define SOCFPGA_CORE_CLK           800U
#define SOCFPGA_PERIPHCLK          (SOCFPGA_CORE_CLK/4U)
#define MPCORE_GTM_CLOCK_FREQ_MHZ  SOCFPGA_PERIPHCLK

/*
 * Tmer preescaler and load value to achieve 1ms per tick.
 * Note that the preescaler value must be representable with 8 bits;
 * and the load value must be a 32bit value.
 *
 * Private timers and watchdogs are clocked with PERIPHCLK which is 1/2 of
 * the CPU frequency. The formula is:
 *
 *    period = (preescaler + 1)(load + 1) / PERIPHCLK
 *      - period    = 1ms
 *      - periphclk = 200MHz
 *
 * For example:
 *      - preescaler = 0
 *      - load       = 200000
 */
#define MPCORE_WDT_PS_1MS    0U                          /* preescaler for private watchdog */
#define MPCORE_WDT_LOAD_1MS  (SOCFPGA_PERIPHCLK * 1000U) /* load value for private watchdog */

#define MPCORE_TM_PS_1MS     0U                          /* preescaler for private timer */
#define MPCORE_TM_LOAD_1MS   (SOCFPGA_PERIPHCLK * 1000U) /* load value for private timer */

#define MPCORE_GTM_PS_1MS     0U                          /* preescaler for global timer */
#define MPCORE_GTM_LOAD_1MS   (SOCFPGA_PERIPHCLK * 1000U) /* load value for global timer */

/*
 * Memory base address and size
 */
#define DDR_ADDR 0x00000000
#define DDR_SIZE 0x40000000 /* 1000MB */

#define OSRAM_ADDR 0xFFFF0000

/*
 *  MPCore Private Memory Region Base Address
 */
#define MPCORE_PMR_BASE  0xfffec000

/*
 *  Number of interrupts supported by the GICv1.0 in this board. Note
 *  that these values could be obtained dynamically from the corresponding
 *  GIC register.
 */
#define DIC_TMIN_INTNO    0U
#define DIC_TMAX_INTNO   211U // maximum interrupt ID
#define DIC_TNUM_INT     212U // maximum number of interrupts

/*
 *
 */
#define H2F_OFFSET    0xC0000000
#define LWH2F_OFFSET  0xff200000

/*
 *  RSTMGR
 */
#define RSTMGR_BASE 0xFFD05000
#define RSTMGR_MPUMODRST  (RSTMGR_BASE + 0x10)
#define RSTMGR_PERMODRST  (RSTMGR_BASE + 0x14)
#define RSTMGR_BRGMODRST  (RSTMGR_BASE + 0x1C)

#define RSTMGR_MPUMODRST_CPU1  0x02

#define RSTMGR_PERMODRST_L4WD0 0x40

#define RSTMGR_BRGMODRST_H2F   0x01
#define RSTMGR_BRGMODRST_LWH2F 0x02 
#define RSTMGR_BRGMODRST_F2H   0x04

/*
 *  SYSMGR
 */
#define SYSMGR_BASE 0xFFD08000
#define SYSMGR_CPU1STARTADDR (SYSMGR_BASE + 0xC4)

/*
 *  FPGAMGR
 */
#define FPGAMGR_BASE 0xFF706000
#define FPGAMGR_CTRL (SYSMGR_BASE + 0x04)
  
/*
 *  L3REGS
 */
#define L3REGS_BASE 0xFF800000
#define L3REGS_REMAP (L3REGS_BASE + 0x00)

#define L3REGS_REMAP_MPUZERO     0x01
#define L3REGS_REMAP_NONMPUZERO  0x02
#define L3REGS_REMAP_HPS2FPGA    0x08
#define L3REGS_REMAP_LWHP2FPGA   0x10

/*
 *  MPUL2
 */
#define MPUL2_BASE 0xFFFEF000

#define MPUL2_ADRFILTER (MPUL2_BASE  + 0xC00)

/*
 *  UART
 */
#define UART0_BASE  0xFFC02000
#define UART1_BASE  0xFFC03000

#define UART0_LCR_VAL 0
#define UART0_DLM_VAL 0
#define UART0_DLL_VAL 0

#define UART1_LCR_VAL 0
#define UART1_DLM_VAL 0
#define UART1_DLL_VAL 0

/*
 *  Interrupt Number
 */
#define DIC_IRQNO_UART0 194
#define DIC_IRQNO_UART1 195


#ifndef TOPPERS_MACRO_ONLY

/*
 *  Trace log definitions (TODO: not tested)
 *
 *  TRACE_GET_TIM returns microseconds and uses the lower part of the
 *  global timer counter (32bit). This means that only around 12.89 seconds
 *  (0xFFFFFFFF/333 us) of trace can be taken until the counter wraps around.
 *
 *  Note: For traces that take more time, comment the TRACE_HW_INIT and
 *  TRACE_GET_TIM. As defined in trace_config.h, get_my_current_time()
 *  will be used instead. This allows taking a very long trace log (up to
 *  49,7 days) but the resolution will be in the milliseconds order.
 */
#ifndef MPCORE_GTC_COUNT_L
#define MPCORE_GTC_COUNT_L (MPCORE_PMR_BASE + 0x0200)
#endif  /* MPCORE_GTC_COUNT_L */

#define TRACE_HW_INIT() x_init_pcc()
#define TRACE_GET_TIM() (*((volatile uint32_t *)(MPCORE_GTC_COUNT_L)) / (MPCORE_GTC_CLOCK_FREQ_MHZ))

#endif /* TOPPERS_MACRO_ONLY */
#endif /* TOPPERS_SOCFPGA_H */
