/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2007-2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  嵭Ԥϡʲ(1)(4)ξ˸¤ꡤܥեȥ
 *  ܥեȥѤΤޤࡥʲƱˤѡʣ
 *  ѡۡʰʲѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *      ޤܥեȥΥ桼ޤϥɥ桼Τʤ
 *      ͳ˴Ťᤫ⡤嵭ԤTOPPERSץȤ
 *      դ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơλŪ
 *  ФŬޤơʤݾڤԤʤޤܥեȥ
 *  ѤˤľŪޤϴŪʤ»˴ؤƤ⡤
 *  Ǥʤ
 *
 *  @(#) $Id: target_config.c 1048 2013-11-13 10:29:14Z ertl-honda $  
 */

#include "kernel_impl.h"
#include <sil.h>
#include "ns16550.h"
#include "pl310.h"

/*
 *  entry point (start.S)
 */
extern void start(void);

/*
 *  PPRC2 boot code(target_support.S)
 */
extern uint8_t prc2_boot;
extern uint8_t prc2_boot_end;

static void
copy_prc2_boot(void){
	int cnt;
	volatile uint8_t *dst_addr;

	dst_addr = (uint8_t*)0x00000000; /* DDR */
	for (cnt = 0; cnt < (&prc2_boot_end - &prc2_boot); cnt++) {
		*(dst_addr + cnt) = *(volatile uint8_t *)(&prc2_boot + cnt);
	}

	__asm__ volatile("dsb");
	__asm__ volatile("isb");
}


/*
 *  Master processor initialization before str_ker().
 *  (E.g.: initialize the interrupt controller or the SCU.)
 */
void
target_mprc_initialize(void)
{
	chip_mprc_initialize();

#ifndef TOPPERS_SAFEG_NONSECURE
	/* map DDR to 0x00000000 - 0xffff0000 (also for uboot)*/
	sil_wrw_mem((uint32_t*)MPUL2_ADRFILTER, 0x01);
#endif  /* TOPPERS_SAFEG_NONSECURE */

#if TNUM_PRCID >= 2

#if !defined(TOPPERS_SAFEG_SECURE) & !defined(TOPPERS_SAFEG_NONSECURE)
	/*
	 * SafeG ̤ѻ
	 */
	/* Core2 boot code to DDR */
	copy_prc2_boot();
	/* Set Core2 boot address */
	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, (uint32_t)start);
	/* Wake up processor 2 */
	sil_wrw_mem((uint32_t*)RSTMGR_MPUMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_MPUMODRST) & ~RSTMGR_MPUMODRST_CPU1);
#endif /* !defined(TOPPERS_SAFEG_SECURE) & !defined(TOPPERS_SAFEG_NONSECURE) */

#ifdef TOPPERS_SAFEG_NONSECURE
	/*
	 *  SafeG ѻ(NonSecure¦) 
	 *  ץå2Trustˤơ
	 *  SYSMGR_CPU1STARTADDR ꤵΤԤäƤ뤿ᡤ
	 *  ư뤿˽񤭹ࡥ
	 */
	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, (uint32_t)start);
#endif /* TOPPERS_SAFEG_NONSECURE */

#endif /* TNUM_PRCID >= 2 */
}

void
target_fput_log(char c)
{
	if (c == '\n') {
		sio_pol_putc('\r', 1);
	}
		sio_pol_putc(c, 1);
}

/*
 * MMU initialization (called from the chip code)
 *
 * We create a simpe 1-to-1 (physical address = virtual address) memory
 * mapping and enable caching for the DDR memory.
 */
void
target_mmu_init(void)
{
	MEMORY_ATTRIBUTE m_attribute;

	/*
	 * Basic initialization of the whole virtual memory space to
	 * non-cachable, non-bufferable, strongly-ordered.
	 */
	m_attribute.pa   = 0x00000000;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = 0x80000000; /* 2GB */
	m_attribute.ns   = 0;          /* 0=Secure */
	m_attribute.s    = 1;          /* 1=Shared */
	m_attribute.ap   = 3;          /* Full access */
	m_attribute.tex  = 0;          /* Strongly Ordered */
	m_attribute.c    = 0;
	m_attribute.b    = 0;

	mmu_map_memory (&m_attribute);

	m_attribute.pa   = 0x80000000; /* 2GB */
	m_attribute.va   = m_attribute.pa;

	mmu_map_memory (&m_attribute);

	/*
	 *  Enable caching for the DDR
	 */
	m_attribute.pa   = DDR_ADDR;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = DDR_SIZE;
	m_attribute.ns   = 0;      /* NonSecure    */
	m_attribute.s    = 1;      /* 1=Shared */
	m_attribute.ap   = 3;      /* Full access */
	/* Outer and Inner Write-Back, Write-Allocate */
	m_attribute.tex  = 1;             
	m_attribute.c    = 1;             
	m_attribute.b    = 1;             

	mmu_map_memory(&m_attribute);

#ifdef TOPPERS_SAFEG_SECURE
	/*
	 *  Enable caching for the DDR
	 */
	m_attribute.pa   = DDR_ADDR;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = 0x3F000000;
	m_attribute.ns   = 1;        /* NonSecure    */
	m_attribute.s    = 1;        /* 1=Shared */
	m_attribute.ap   = 3;        /* Full access */

	/*
	 * ǤʤȡL2νΤLinuxidݤSafeG
	 * Linuxѿ˽񤭹ͤȿǤʤ.
	 * L2ͭͭ׸Ƥ
	 */
	
	/* åѿξϤư */
	/* Onter and Inner Write-Back; no Write-Allocate ng */ 
	m_attribute.tex  = 0;
	m_attribute.c    = 1;
	m_attribute.b    = 1;

	/* ХѿξϤư */
	/* Outer and Inner Write-Through, no Write-Allocate */
//	m_attribute.tex  = 0;
//	m_attribute.c    = 1;
//	m_attribute.b    = 0;

	mmu_map_memory(&m_attribute);

	/*
	 * SafeG-COM shared memory area
	 */
	m_attribute.pa   = SAFEG_COM_MEM_ADDR;
	m_attribute.va   = m_attribute.pa;
	m_attribute.size = SAFEG_COM_MEM_SIZE * 1024 * 1024;
	m_attribute.ns   = 1;          /* NonSecure    */
	m_attribute.s    = 1;          /* ͭ     */
	m_attribute.ap   = 3;          /* ե륢 */
	m_attribute.tex  = 0;          /* åON */
	m_attribute.c    = 1;          /* Outer and Inner */          
	m_attribute.b    = 1;          /* Inner Write-Back, Write Allocate */

	mmu_map_memory(&m_attribute);
#endif /* TOPPERS_SAFEG_SECURE */
}

/*
 *  Vector Tablechip_support.S
 */
extern void *vector_table;

/*
 *  Disable L4WD0
 */
static void
l4wd0_disable(void)
{
	/* reset watchdog  */ 
	sil_wrw_mem((uint32_t*)RSTMGR_PERMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_PERMODRST) | RSTMGR_PERMODRST_L4WD0);
	/* clear reset watchdog  */
	sil_wrw_mem((uint32_t*)RSTMGR_PERMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_PERMODRST) & ~RSTMGR_PERMODRST_L4WD0);
}

/*
 *  Enable Bridge
 */
static bool_t
enable_bridges(void)
{
	/* reset bridges  */ 
	sil_wrw_mem((uint32_t*)RSTMGR_BRGMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_BRGMODRST) |
				(RSTMGR_BRGMODRST_H2F|RSTMGR_BRGMODRST_LWH2F|RSTMGR_BRGMODRST_F2H));

	/* ToDo error check */

	/* clear reset bridges  */ 
	sil_wrw_mem((uint32_t*)RSTMGR_BRGMODRST,
				sil_rew_mem((uint32_t*)RSTMGR_BRGMODRST) &
				~(RSTMGR_BRGMODRST_H2F|RSTMGR_BRGMODRST_LWH2F|RSTMGR_BRGMODRST_F2H));

	/* Map hps2fpga and lwhp2fpga */
	sil_wrw_mem((uint32_t*)L3REGS_REMAP,
				sil_rew_mem((uint32_t*)L3REGS_REMAP) | (L3REGS_REMAP_HPS2FPGA|L3REGS_REMAP_LWHP2FPGA));

	return true;
}

/*
 *  L2åν
 */
static void l2cache_init(void)
{
#ifdef TOPPERS_SAFEG_NONSECURE
	char l2init_id_syscall_name[] = "l2init";
	uint32_t l2init_id = 0;
	uint32_t err;

	err = safeg_syscall_getid(l2init_id_syscall_name, &l2init_id);
	safeg_syscall_invoke(l2init_id, 0, 0, 0);
#else /* !TOPPERS_SAFEG_NONSECURE */
	uint32_t aux_ctrl = 0;

	aux_ctrl |= (1 << L2X0_AUX_CTRL_SHARE_OVERRIDE_SHIFT) |
	  (1 << L2X0_AUX_CTRL_DATA_PREFETCH_SHIFT) |
		(1 << L2X0_AUX_CTRL_INSTR_PREFETCH_SHIFT);

	pl310_init(aux_ctrl, ~0x0U);
#endif /* TOPPERS_SAFEG_NONSECURE */
}

#ifdef TOPPERS_SAFEG_SECURE
#include "syscalls_api.h"

uint32_t
safeg_l2cache_init(const uint32_t core_id,
				   const uint32_t ns,
				   const uint32_t a,
				   const uint32_t b,
				   const uint32_t c)
{
	l2cache_init();
	return SAFEG_OK(SAFEG_SYSCALL_RET__DONT_SWITCH);
}

struct safeg_syscall l2init_call = {
	.is_t_callable  = 1,
	.is_nt_callable = 1,
	.name = "l2init",
	.function = safeg_l2cache_init,
};

/*
 *  L2åǥХå쥸񤭹
 */
uint32_t
l2cache_debug_set(const uint32_t core_id,
				  const uint32_t ns,
				  const uint32_t a,
				  const uint32_t b,
				  const uint32_t c)
{
	pl310_debug_set(a);
	return SAFEG_OK(SAFEG_SYSCALL_RET__DONT_SWITCH);
}

struct safeg_syscall l2debug_set_call = {
	.is_t_callable  = 1,
	.is_nt_callable = 1,
	.name = "l2debug",
	.function = l2cache_debug_set,
};

#endif /* TOPPERS_SAFEG_SECURE */

#if !defined(TOPPERS_SAFEG_SECURE)
volatile uint32_t target_initialize_end[TNUM_PRCID];
#endif /* !TOPPERS_SAFEG_SECURE */

void
target_initialize(void)
{
	/* Disbale high vector */
	set_high_vector(false);

	/* Set Vector Table */
	CP15_SET_VBAR((uint32_t) &vector_table);

#ifndef TOPPERS_SAFEG_NONSECURE
	if (x_sense_mprc()) {
		/* Disable Watchdog(L4WD0) */
		l4wd0_disable();
		/* Enable FPGA bridge */
		enable_bridges();
	}
#endif /*  TOPPERS_SAFEG_NONSECURE */

#ifdef TOPPERS_SAFEG_SECURE
	/* Set NULL(0x00) for Linux second core boot */
	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, 0x00);
#endif /* TOPPERS_SAFEG_SECURE */

	/*
	 * Initialize chip
	 */
	chip_initialize();

#ifdef TOPPERS_SAFEG_SECURE
	if (x_sense_mprc()) {
		/*
		 * L2åؿSafeGϿ
		 */
		uint32_t id;
		safeg_syscall_regdyn(&l2init_call, &id);

		/*
		 * L2åǥХå쥸񤭹ߴؿSafeGϿ
		 */
		safeg_syscall_regdyn(&l2debug_set_call, &id);
	}
#endif /*  TOPPERS_SAFEG_SECURE */

	/*
	 *  Initialize the UART. If we are using a
	 *  global logger (shared by all cores) only
	 *  the master processor (mprc) initializes it.
	 */
#ifdef G_SYSLOG
	if (x_sense_mprc()) {
	}
#else
	
#endif /* G_SYSLOG */

#if !defined(TOPPERS_SAFEG_SECURE)
	/* Enable L2 cache by myself in case of without safeg or with safeg Non-Secure */
	if (x_sense_mprc()) {
		uint32_t i;
		uint32_t sum = 0;
		while(1) {
			for(i = 1; i < TNUM_PRCID; i++) {
				sum += target_initialize_end[i];
			}
			if (sum == (TNUM_PRCID - 1)) {
				break;
			}
		}
		l2cache_init();
	}
	target_initialize_end[x_prc_index()] = 1;
#endif /* !TOPPERS_SAFEG_SECURE */
}

/*
 *  Target-dependent exit processing
 */
void
target_exit(void)
{
	/*
	 *  Chip exit processing
	 */
	chip_exit();

	/*
	 *  Endless loop
	 */
	while(1);
}


#ifdef TOPPERS_SAFEG_SECURE

/*
 *  Return Non-Trust Start address.
 *  If Non-Trust not ready, return -1;
 */
uint32_t
target_get_nt_start_address(void) {
	uint32_t addr;

	if (x_sense_mprc()) {
		return TARGET_PRC1_NS_START_ADDR;
	}
	else {
		addr = sil_rew_mem((void*)SYSMGR_CPU1STARTADDR);
		if(addr == 0x00) {
			return -1;
		}
		return addr;
	}
}

#define ZIMAGE_SRC   0x000008000
#define ZIMAGE_SIZE  (5*1024*1024)
#define ZIMAGE_DEST  0x3f400000

#define DTB_SRC   0x00000100
#define DTB_SIZE  (512*1024)
#define DTB_DEST  (ZIMAGE_DEST + ZIMAGE_SIZE)

#define UBOOT_SRC   0x01000040
#define UBOOT_SIZE  (1*1024*1024)
#define UBOOT_DEST  (ZIMAGE_DEST + ZIMAGE_SIZE + DTB_SIZE)

/*
 *  Prepare Reboot Non-Trust Linux
 */
void
target_prepare_reboot(void){
	volatile uint32_t *src;
	volatile uint32_t *dest;
	int i;

	src = (volatile uint32_t *)ZIMAGE_SRC;
	dest = (volatile uint32_t *)ZIMAGE_DEST;
	syslog(LOG_EMERG, "backup zimage from 0x%x to 0x%x",src, dest);
	for(i = 0;  i < ZIMAGE_SIZE/4; i++) {
		*dest++ = *src++;
	}

	src = (volatile uint32_t *)DTB_SRC;
	dest = (volatile uint32_t *)DTB_DEST;
	syslog(LOG_EMERG, "backup dtb from 0x%x to 0x%x",src, dest);
	for(i = 0;  i < DTB_SIZE/4; i++) {
		*dest++ = *src++;
	}

	src = (volatile uint32_t *)UBOOT_SRC;
	dest = (volatile uint32_t *)UBOOT_DEST;
	syslog(LOG_EMERG, "backup u-boot from 0x%x to 0x%x",src, dest);
	for(i = 0;  i < UBOOT_SIZE/4; i++) {
		*dest++ = *src++;
	}
}

/*
 *  Reboot Non-Trust Linux
 */
void
target_do_reboot(void){
	int i;
	volatile uint32_t *src;
	volatile uint32_t *dest;
	uint32_t ret;

	src = (volatile uint32_t *)ZIMAGE_DEST;
	dest = (volatile uint32_t *)ZIMAGE_SRC;
	for(i = 0;  i < ZIMAGE_SIZE/4; i++) {
		*dest++ = *src++;
	}

	src = (volatile uint32_t *)DTB_DEST;
	dest = (volatile uint32_t *)DTB_SRC;
	for(i = 0;  i < DTB_SIZE/4; i++) {
		*dest++ = *src++;
	}

	src = (volatile uint32_t *)UBOOT_DEST;
	dest = (volatile uint32_t *)UBOOT_SRC;
	for(i = 0;  i < UBOOT_SIZE/4; i++) {
		*dest++ = *src++;
	}

	/* Då奯ꥢ */
	mpcore_dcache_clean();

	ret = safeg_syscall_restarnt();
	if (ret != SAFEG_SYSCALL_ERROR__OK) {
		syslog(LOG_NOTICE, "SafeG ERROR: %u", ret);
	}

	/* L2̵ */
	pl310_disable();

	sil_wrw_mem((uint32_t*)SYSMGR_CPU1STARTADDR, 0x00);
}

#endif /* TOPPERS_SAFEG_SECURE */
